// VZ Image Compare (C) 2020 Stephane Charette <stephanecharette@gmail.com>
// MIT license applies.  See "license.txt" for details.
// $Id: main.cpp 2966 2020-05-04 08:02:14Z stephane $


#include <iostream>
#include <sstream>
#include "VzImgCmp.hpp"


cv::Mat get_image(const std::string & filename)
{
	std::cout << "-> reading " << filename << ": ";

	if (filename.empty())
	{
		throw std::invalid_argument("cannot read mage from empty filename");
	}

	cv::Mat mat = cv::imread(filename, cv::IMREAD_COLOR);

	if (mat.channels() != 3 ||
		mat.cols < 10 ||
		mat.rows < 10 )
	{
		throw std::runtime_error("failed to read image " + filename);
	}

	std::cout << mat.cols << "x" << mat.rows << std::endl;

	return mat;
}


void run_test_1(int argc, char *argv[])
{
	vz::ImgCmp image_compare;

	std::cout
		<< "vz::ImgCmp version .. " << image_compare.version()	<< std::endl
		<< "OpenCV version ...... " << CV_VERSION				<< std::endl;

	image_compare.resized_image_scale	= 0.25;
	image_compare.dilate_and_erode		= 3;
	image_compare.min_contour_area		= 2.0;

//	image_compare.set_flag	(vz::ImgCmp::Flags::kDiffColour		);
	image_compare.set_flag	(vz::ImgCmp::Flags::kDiffGreyscale	);

	image_compare.set_flag	(vz::ImgCmp::Flags::kThresholdTriangle);
//	image_compare.set_flag	(vz::ImgCmp::Flags::kThresholdOtsu	);

	image_compare.set_flag	(vz::ImgCmp::Flags::kDiffOriginalSize);
//	image_compare.set_flag	(vz::ImgCmp::Flags::kDiffResized	);

//	image_compare.set_flag(vz::ImgCmp::Flags::kAnnotateAddRedBorder);
	image_compare.clear_flag(vz::ImgCmp::Flags::kAnnotateAddRedBorder);

//	image_compare.set_flag(vz::ImgCmp::Flags::kAnnotateOverColour);
	image_compare.clear_flag(vz::ImgCmp::Flags::kAnnotateOverColour);

	image_compare.set_flag	(vz::ImgCmp::Flags::kDrawContour	);
//	image_compare.clear_flag(vz::ImgCmp::Flags::kDrawContour	);

	image_compare.set_flag	(vz::ImgCmp::Flags::kDrawRectangle	);
//	image_compare.clear_flag(vz::ImgCmp::Flags::kDrawRectangle	);

	image_compare.set_master_image(get_image(argv[1]));
	cv::Mat resized = image_compare.master_resized.get();

	cv::namedWindow(	"vz::ImgCmp - master image (resized)", cv::WINDOW_KEEPRATIO + cv::WINDOW_GUI_EXPANDED);
	cv::resizeWindow(	"vz::ImgCmp - master image (resized)", resized.cols, resized.rows);
	cv::imshow(			"vz::ImgCmp - master image (resized)", resized);

	for (int idx = 2; idx < argc; idx ++)
	{
		image_compare.compare(get_image(argv[idx]));
		image_compare.annotate();
//		image_compare.debug_save();

		cv::namedWindow(	"vz::ImgCmp - candy (resized)", cv::WINDOW_KEEPRATIO + cv::WINDOW_GUI_EXPANDED);
		cv::resizeWindow(	"vz::ImgCmp - candy (resized)", resized.cols, resized.rows);
		cv::imshow(			"vz::ImgCmp - candy (resized)", image_compare.candy_resized.get());

		cv::namedWindow(	"vz::ImgCmp - differences (resized)", cv::WINDOW_KEEPRATIO + cv::WINDOW_GUI_EXPANDED);
		cv::resizeWindow(	"vz::ImgCmp - differences (resized)", image_compare.differences_threshold.cols, image_compare.differences_threshold.rows);
		cv::imshow(			"vz::ImgCmp - differences (resized)", image_compare.differences_threshold);

		cv::namedWindow(	"vz::ImgCmp - annotation (resized)", cv::WINDOW_KEEPRATIO + cv::WINDOW_GUI_EXPANDED);
		cv::resizeWindow(	"vz::ImgCmp - annotation (resized)", image_compare.annotated_candy.cols, image_compare.annotated_candy.rows);
		cv::imshow(			"vz::ImgCmp - annotation (resized)", image_compare.annotated_candy);

		const int key = cv::waitKeyEx(-1);
		if (key == 1048603) // <ESC>
		{
			break;
		}
	}

	return;
}


void run_test_2(int argc, char *argv[])
{
	vz::ImgCmp image_compare;

	std::cout
		<< "vz::ImgCmp version .. " << image_compare.version()	<< std::endl
		<< "OpenCV version ...... " << CV_VERSION				<< std::endl;

	image_compare.dilate_and_erode		= 3;
	image_compare.resized_image_scale	= 0.5;
	image_compare.min_contour_area		= 20.0;

//	image_compare.set_flag	(vz::ImgCmp::Flags::kDiffColour				);
	image_compare.set_flag	(vz::ImgCmp::Flags::kDiffGreyscale			);
	image_compare.set_flag	(vz::ImgCmp::Flags::kThresholdTriangle		);
//	image_compare.set_flag	(vz::ImgCmp::Flags::kDiffOriginalSize		);
	image_compare.set_flag	(vz::ImgCmp::Flags::kDiffResized			);
	image_compare.set_flag	(vz::ImgCmp::Flags::kAnnotateAddRedBorder	);
	image_compare.set_flag	(vz::ImgCmp::Flags::kAnnotateAddGreenBorder	);
//	image_compare.set_flag	(vz::ImgCmp::Flags::kAnnotateOverColour		);
	image_compare.set_flag	(vz::ImgCmp::Flags::kAnnotateOverGrey		);
	image_compare.set_flag	(vz::ImgCmp::Flags::kDrawContour			);
	image_compare.set_flag	(vz::ImgCmp::Flags::kDrawRectangle			);

	const int width = 900;

	cv::Mat tmp = get_image(argv[1]);
	image_compare.set_master_image(tmp);
	cv::Mat master_image = vz::resize(tmp, width);

	cv::namedWindow(	"vz::ImgCmp - master image", cv::WINDOW_KEEPRATIO + cv::WINDOW_GUI_EXPANDED);
	cv::resizeWindow(	"vz::ImgCmp - master image", master_image.cols, master_image.rows);
	cv::imshow(			"vz::ImgCmp - master image", master_image);

	for (int idx = 2; idx < argc; idx ++)
	{
		cv::Mat comparison_image = get_image(argv[idx]);
		image_compare.compare(comparison_image);

		comparison_image = vz::resize(comparison_image, width);
		cv::namedWindow(	"vz::ImgCmp - comparison image", cv::WINDOW_KEEPRATIO + cv::WINDOW_GUI_EXPANDED);
		cv::resizeWindow(	"vz::ImgCmp - comparison image", comparison_image.cols, comparison_image.rows);
		cv::imshow(			"vz::ImgCmp - comparison image", comparison_image);

		cv::Mat annotation_image = image_compare.annotate();
		annotation_image = vz::resize(annotation_image, width);
		cv::namedWindow(	"vz::ImgCmp - annotation image", cv::WINDOW_KEEPRATIO + cv::WINDOW_GUI_EXPANDED);
		cv::resizeWindow(	"vz::ImgCmp - annotation image", annotation_image.cols, annotation_image.rows);
		cv::imshow(			"vz::ImgCmp - annotation image", annotation_image);

		cv::Mat mask = image_compare.differences_threshold;
		mask = vz::resize(mask, width);
		cv::namedWindow(	"vz::ImgCmp - mask", cv::WINDOW_KEEPRATIO + cv::WINDOW_GUI_EXPANDED);
		cv::resizeWindow(	"vz::ImgCmp - mask", mask.cols, mask.rows);
		cv::imshow(			"vz::ImgCmp - mask", mask);

		cv::namedWindow(	"vz::ImgCmp - flip between master and comparison", cv::WINDOW_KEEPRATIO + cv::WINDOW_GUI_EXPANDED);
		cv::resizeWindow(	"vz::ImgCmp - flip between master and comparison", master_image.cols, master_image.rows);

		int key = -1;
		size_t count = 0;
		while (key == -1)
		{
			count ++;

			if (count % 2)
			{
				cv::imshow("vz::ImgCmp - flip between master and comparison", master_image);
			}
			else
			{
				cv::imshow("vz::ImgCmp - flip between master and comparison", comparison_image);
			}

			key = cv::waitKeyEx(333);
		}
		if (key == 1048603) // <ESC>
		{
			break;
		}
	}

	return;
}


int main(int argc, char *argv[])
{
	int result = 1;

	try
	{
		if (argc < 3)
		{
			std::cout
				<< ""																		<< std::endl
				<< "Usage:"																	<< std::endl
				<< ""																		<< std::endl
				<< "\t" << argv[0] << " <master image> <test image 1> [<test image 2>...]"	<< std::endl
				<< ""																		<< std::endl;
		}
		else
		{
//			run_test_1(argc, argv);
			run_test_2(argc, argv);
		}

		result = 0;
	}
	catch (const std::exception & e)
	{
		std::cout << "ERROR:" << std::endl << e.what() << std::endl;
		result = 2;
	}
	catch (...)
	{
		std::cout << "ERROR:" << std::endl << "unknown exception caught -- exiting" << std::endl;
		result = 3;
	}

	return result;
}
